import unittest

from lspace import create_app


class TestCreateApp(unittest.TestCase):
    def test_create_app(self):
        """
        basic first test: just check if it breaks when app is created and commands imported

        :return:
        """
        app = create_app()
        from lspace.cli import cli as cli_group
        assert app



import datetime

from sqlalchemy import Column, String, DateTime, Text, Integer

from lspace import db

import json

class MetaCache(db.Model):
    __tablename__ = 'meta_cache'

    id = Column(Integer, primary_key=True)
    isbn = Column(String(13))
    service = Column(String(10))
    date = Column(DateTime(), default=datetime.datetime.now())

    # sqlite doesnt like json, and we dont want to run queries on json keys,
    # so we just use text and parse.
    data = Column(Text())

    @property
    def results(self):
        results = json.loads(self.data)
        return results

    @results.setter
    def results(self, results):
        data = json.dumps(results)
        self.data = data



__version__ = '0.1.12'

import os

APP_NAME = 'lspace'
os.environ['FLASK_APP'] = os.path.join(os.path.dirname(__file__), 'app.py')

from lspace.helpers.init_logging import init_logging
init_logging('debug')
#print(app.config['USER_CONFIG'].get('loglevel', 'info'))

import click
from flask import Flask
from flask_marshmallow import Marshmallow
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy
from flask_whooshee import Whooshee


db = SQLAlchemy()
migrate = Migrate()
whooshee = Whooshee()
marshmallow = Marshmallow()


def create_app(config_path=None, app_dir=None):
    from .helpers import read_config

    app = Flask(__name__)

    if not app_dir:
        app_dir = click.get_app_dir(APP_NAME)
    app.config['APP_DIR'] = app_dir

    if not config_path:
        config_path = os.path.join(app.config['APP_DIR'], 'config.yaml')

    app.config['CONFIG_PATH'] = config_path
    app.config['USER_CONFIG'] = read_config(config_path, app_dir)
    app.config['SQLALCHEMY_DATABASE_URI'] = app.config['USER_CONFIG']['database_path']
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False
    app.config['WHOOSHEE_DIR'] = os.path.join(app.config['APP_DIR'], 'whoosh_index')

    app.config['USER_CONFIG']['library_path'] = os.path.abspath(os.path.expanduser(app.config['USER_CONFIG']['library_path']))

    migration_dir = os.path.join(os.path.dirname(__file__), 'migrations')

    db.init_app(app)
    migrate.init_app(app, db, directory=migration_dir)
    whooshee.init_app(app)
    marshmallow.init_app(app)

    return app



import os

from ._base import FileTypeBase
from .pdf import PDF
from .epub import Epub

mapping = {
    '.pdf': PDF,
    '.epub': Epub
}


def get_file_type_object(path):
    filename, file_extension = os.path.splitext(path)

    file_class = mapping.get(file_extension, None)
    if file_class:
        return file_class(path)
    else:
        return None



import os

from lspace.cli import list_command
from tests.cli._cli_test_base import BaseCliTest


class TestList(BaseCliTest):

    def test_list_command_simple(self):
        runner = self.app.test_cli_runner()

        result = runner.invoke(list_command._list, [])

        assert result.exit_code == 0

        print(result.output)

        assert '{author} - {title}'.format(author=self.test_author_name, title=self.test_title) in result.output

    def test_list_command_details(self):
        runner = self.app.test_cli_runner()

        result = runner.invoke(list_command._list, ['--details'])

        assert result.exit_code == 0

        assert '{author} - {title}'.format(author=self.test_author_name, title=self.test_title) in result.output
        assert 'language: ' in result.output
        assert 'year: ' in result.output
        assert 'isbn: ' in result.output

    def test_list_command_path(self):
        runner = self.app.test_cli_runner()

        result = runner.invoke(list_command._list, ['--path'])

        assert result.exit_code == 0

        assert os.path.join(self.test_dir, self.test_path) in result.output




from .. import db

book_author_association_table = db.Table('book_author_association',
                                         db.Column('book_id', db.Integer,
                                                   db.ForeignKey('books.id')),
                                         db.Column('author_id', db.Integer,
                                                   db.ForeignKey('authors.id'))
                                         )



import isbnlib

from lspace import db
from lspace.helpers import logger
from lspace.models.meta_cache import MetaCache
from lspace.models import Book, Author


def _get_metadata_for_isbn(isbn, service='openl'):
    # type: (str, str) -> Book
    cached_meta = MetaCache.query.filter_by(isbn=isbn, service=service).first()
    if cached_meta:
        meta = cached_meta.results
    else:
        try:
            meta = isbnlib.meta(isbn, service=service, cache='default')
        except (isbnlib.dev._exceptions.NoDataForSelectorError,
                isbnlib._exceptions.NotValidISBNError,
                isbnlib.dev._exceptions.DataNotFoundAtServiceError
                ):
            meta = {}
        new_meta = MetaCache(isbn=isbn, service=service, results=meta)

        db.session.add(new_meta)
        db.session.commit()

    if meta:
        return Book.from_search_result(meta, metadata_source=service)
    return None


def query_isbn_data(isbn_str):
    # type: (str) -> Book

    if isbnlib.is_isbn10(isbn_str):
        isbn_str = isbnlib.to_isbn13(isbn_str)

    logger.info('query openlibrary for %s' % isbn_str)
    meta = _get_metadata_for_isbn(isbn_str, 'openl')

    if not meta:
        logger.info('query google books for %s' % isbn_str)
        meta = _get_metadata_for_isbn(isbn_str, 'goob')

    if meta:
        return meta
    return None


def query_google_books(words):
    # type: (str) -> [SearchResult]
    logger.debug('query google books for %s' % words)
    try:
        raw_results = isbnlib.goom(words)

    except isbnlib.dev._exceptions.NoDataForSelectorError as e:
        raw_results = []
    return [Book.from_search_result(result, metadata_source='google books api') for result in raw_results]


def query_db(query, books=True, authors=True):
    if not query or (len(query) == 1 and not query[0]):  # in case of ('', )
        return Book.query.all()
    else:
        results = []
        joined_query = ' '.join(query)
        if books:
            results = Book.query.whooshee_search(joined_query, match_substrings=False).all()
        if authors:
            author_results = Author.query.whooshee_search(joined_query, match_substrings=False).all()
            for author in author_results:
                for book in author.books:
                    results.append(book)
    return results



from lspace.cli.tools_command import update_paths as update_paths_module

from lspace.file_types import FileTypeBase
from lspace.models import Book
from tests.cli._cli_test_base import BaseCliTest

from unittest.mock import MagicMock


class TestToolsUpdatePathsCommand(BaseCliTest):

    def test_update_path(self):
        with self.app.app_context():
            book = Book.query.first()

            test_new_path = 'new_path'
            file_type_object_mock = FileTypeBase(book.path)

            update_paths_module._copy_to_library = MagicMock(return_value=test_new_path)
            update_paths_module.get_file_type_object = MagicMock(return_value=file_type_object_mock)

            new_path = update_paths_module.update_path(book)

            assert test_new_path == new_path
            assert book.path == new_path


    def test_update_paths(self):
        runner = self.app.test_cli_runner()

        result = runner.invoke(update_paths_module._update_paths)

        assert result.exit_code == 0

        assert 'moved ' in result.output




"""empty message

Revision ID: b142d6d9e8da
Revises: 
Create Date: 2019-05-11 14:42:35.971140

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b142d6d9e8da'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('authors',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=100), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('books',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('title', sa.String(length=100), nullable=True),
    sa.Column('isbn13', sa.String(length=13), nullable=True),
    sa.Column('publisher', sa.String(length=100), nullable=True),
    sa.Column('year', sa.Integer(), nullable=True),
    sa.Column('language', sa.String(length=20), nullable=True),
    sa.Column('md5sum', sa.String(length=32), nullable=True),
    sa.Column('path', sa.String(length=400), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('book_author_association',
    sa.Column('book_id', sa.Integer(), nullable=True),
    sa.Column('author_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['author_id'], ['authors.id'], ),
    sa.ForeignKeyConstraint(['book_id'], ['books.id'], )
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('book_author_association')
    op.drop_table('books')
    op.drop_table('authors')
    # ### end Alembic commands ###



import os

import click
import yaml

from lspace.helpers import get_default_config
from lspace.cli import cli


@cli.command(help='generate a new config')
def init():
    from flask import current_app

    app_dir = current_app.config['APP_DIR']
    config_path = current_app.config['CONFIG_PATH']

    if not os.path.isdir(app_dir):
        os.makedirs(app_dir)

    if os.path.exists(config_path):
        if not click.confirm('config exists - override?'):
            return

    default_config = get_default_config(app_dir=app_dir)
    with open(config_path, 'w') as config:
        yaml.dump(default_config, config)
    click.echo('written config file to %s' % config_path)

    return config_path



import logging
import os

from flask import current_app
from slugify import slugify
from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from lspace import db, whooshee
from lspace.models import Author

logger = logging.getLogger(__name__)


@whooshee.register_model('title', 'language', 'isbn13')
class Book(db.Model):
    __tablename__ = 'books'

    id = Column(Integer, primary_key=True)
    title = Column(String(100))
    authors = relationship("Author",
                           secondary="book_author_association",
                           backref="authors_books")
    isbn13 = Column(String(13))
    publisher = Column(String(100))
    year = Column(Integer())
    language = Column(String(20))

    md5sum = Column(String(32))
    path = Column(String(400))

    metadata_source = Column(String(20), default='')

    @property
    def full_path(self):
        # type: () -> str
        library_path = current_app.config['USER_CONFIG']['library_path']
        return os.path.join(library_path, self.path)

    @property
    def authors_names(self):
        # type: () -> str
        return ', '.join(author.name for author in self.authors)

    @property
    def author_names_slug(self):
        author_slugs = [slugify(author.name) for author in self.authors]
        authors = '_'.join(author_slugs)
        return authors

    @property
    def title_slug(self):
        return slugify(self.title)

    @property
    def extension(self):
        # type: () -> str
        filename, file_extension = os.path.splitext(self.full_path)
        return file_extension

    @staticmethod
    def from_search_result(d, metadata_source):
        book = Book()
        book.metadata_source = metadata_source
        book.from_dict(d)
        return book

    def from_dict(self, d):
        self.isbn13 = d.get('ISBN-13', None)
        self.title = d.get('Title', 'no title')
        self.publisher = d.get('Publisher', None)
        self.year = int(d.get('Year', 'no year'))
        self.language = d.get('Language', None)

        if not d.get('Authors', None):
            authors = ['no author']
        else:
            authors = d.get('Authors')

        for author_name in authors:
            author = Author.query.filter_by(name=author_name).first()
            if not author:
                logger.info('creating %s' % author_name)
                author = Author(name=author_name)
            self.authors.append(author)

    def __repr__(self):
        return '<isbn={isbn} authors={authors} title={title}>'.format(isbn=self.isbn13, authors=self.authors_names,
                                                                      title=self.title)

    def to_dict(self):
        return dict(
            title=self.title,
            authors=self.authors,

            isbn13=self.isbn13,
            publisher=self.publisher,
            year=self.year,
            language=self.language,

            md5sum=self.md5sum,
            path=self.path,

            metadata_source=self.metadata_source
        )

    def formatted_output_head(self):
        return '{authors} - {title} ({year})'.format(
            authors=self.authors_names, title=self.title, year=self.year)

    def formatted_output_details(self):
        return 'isbn: {isbn}\npublisher: {publisher}\nlanguage: {language}\nmetadata source: {source}\n'.format(
            isbn=self.isbn13,
            language=self.language,
            publisher=self.publisher,
            source=self.metadata_source)

    def save(self):
        print(self)
        print('self id', self.id)

        db.session.add(self)

        return db.session.commit()



import logging
import logging.config

def init_logging(loglevel='info'):

    logging.config.dictConfig({
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'default': {
                'format': '%(name)s [%(levelname)s]: %(message)s'
            },
        },
        'handlers': {
            'default': {
                'level': loglevel.upper(),
                'class': 'logging.StreamHandler',
                'formatter': 'default'
            },
        },
        'loggers': {
            '': {
                'handlers': ['default'],
                'level': loglevel.upper(),
                'propagate': True
            },
            'sqlalchemy.engine': {
                'handlers': ['default'],
                'level': loglevel.upper(),
                'propagate': False
            }
        }
    })
    logging.debug('{loglevel} logging!'.format(loglevel=loglevel))




import ebooklib
import html2text
import isbnlib
from ebooklib import epub

from ._base import FileTypeBase


class Epub(FileTypeBase):
    extension = '.epub'

    def __init__(self, path):
        super().__init__(path)
        self.book = epub.read_epub(path)

    def get_text(self):
        # type: () -> [str]
        text = []

        for doc in self.book.get_items_of_type(ebooklib.ITEM_DOCUMENT):
            content = doc.get_content().decode('utf-8')
            text.append(html2text.html2text(content))
        if True in [bool(t) for t in text]:
            return text
        return []

    def get_author(self):
        _author = self.book.get_metadata('DC', 'creator')
        # [('Firstname Lastname ', {})]
        if _author:
            return _author[0][0]
        return None

    def get_title(self):
        _title = self.book.get_metadata('DC', 'title')
        # [('Ratio', {})]
        if _title:
            return _title[0][0]
        return None

    def get_isbn(self):
        _isbn = self.book.get_metadata('DC', 'identifier')
        # [('Ratio', {})]
        if _isbn:
            isbn = _isbn[0][0]
            if not isbn:
                return None
            if isbnlib.notisbn(isbn):
                return None
            if isbnlib.is_isbn10(isbn):
                return isbnlib.to_isbn13(isbn)
            return isbn
        return None



import logging
import os

import click
import isbnlib
import yaml

from lspace import APP_NAME

logger = logging.getLogger(__name__)


def get_default_config(app_dir):
    this_dir = os.path.dirname(os.path.abspath(__file__))
    path_from_here = '../config/'
    with open(os.path.join(this_dir, path_from_here, 'default_conf.yaml'), 'r') as default_config_file:
        default_config = yaml.load(default_config_file, Loader=yaml.SafeLoader)
        default_config['database_path'] = default_config['database_path'].format(
            APP_DIR=app_dir)
    return default_config


def read_config(config_path, app_dir):
    if os.path.isfile(config_path):
        with open(config_path, 'r') as config:
            conf = yaml.load(config, Loader=yaml.SafeLoader)

    else:
        conf = {}
    
    config = get_default_config(app_dir)
    config.update(conf)
    return config

def find_unused_path(base_path, book_path_format, authors, title, extension):
    # type: (str, str, str, str, str) -> str
    """

    :param base_path: 
    :param book_path_format: 
    :param authors: 
    :param title: 
    :param extension: 
    :return: new path relative from base_path  
    """
    # create the path for the book

    count = 0

    while count < 100:
        path_from_base_path = book_path_format.format(
            AUTHORS=authors, TITLE=title)
        # if, for some reason, the path starts with /, we need to make it relative
        while path_from_base_path.startswith(os.sep):
            logger.debug('trimming path to %s' % path_from_base_path[1:])
            path_from_base_path = path_from_base_path[1:]

        if count == 0:
            path_from_base_path += extension
        else:
            path_from_base_path = '{path_from_base_path}_{count}{extension}'.format(
                path_from_base_path=path_from_base_path,
                count=count, extension=extension)

        target_path = os.path.join(base_path, path_from_base_path)

        if not os.path.exists(target_path):
            return path_from_base_path

        count += 1
    return False


def preprocess_isbns(isbns):
    """

    :param isbns: isbns in different formats
    :return: canonical isbn13s
    """
    canonical_isbns = []
    for isbn in isbns:
        if not isbnlib.notisbn(isbn, level='strict'):
            if isbnlib.is_isbn10(isbn):
                isbn = isbnlib.to_isbn13(isbn)
            isbn = isbnlib.get_canonical_isbn(isbn)
            canonical_isbns.append(isbn)
    canonical_isbns = set(canonical_isbns)
    return list(canonical_isbns)



import hashlib
import logging
import os
import string

import isbnlib

from lspace.helpers import preprocess_isbns
from lspace.helpers.query import query_isbn_data, query_google_books
from flask import copy_current_request_context

logger = logging.getLogger(__name__)


class FileTypeBase:
    extension = None

    def __init__(self, path):
        self.path = path

    def get_text(self):
        raise NotImplementedError

    def get_title(self):
        return None

    def get_author(self):
        return None

    def get_year(self):
        return None

    def get_isbn(self):
        return None

    @property
    def filename(self):
        filename = os.path.split(self.path)[-1]
        return filename

    def get_md5(self):
        with open(self.path, 'rb') as file_to_check:
            data = file_to_check.read()
            md5sum = hashlib.md5(data).hexdigest()
        logger.debug('md5 is %s' % md5sum)
        return md5sum

    def find_isbn_in_metadata(self):
        _isbn = self.get_isbn()
        if _isbn:
            logger.info('found isbn %s in metadata!' % _isbn)
            d = query_isbn_data(_isbn)
            if d:
                return [d]
        return []

    def find_isbn_in_text(self):
        logger.info('looking for isbn in text...')
        isbns = self.get_isbns_from_text()
        if isbns:
            isbns_with_metadata = []
            for isbn in isbns:
                d = query_isbn_data(isbn)
                if d:
                    isbns_with_metadata.append(d)

            if isbns_with_metadata:
                logger.info('found isbns in text!')
                return isbns_with_metadata
        return []

    def find_isbn_from_author_title(self):
        if self.get_title():
            search_str = self._filter_symbols(self.get_title())
            if self.get_author():
                search_str = '%s %s' % (self._filter_symbols(self.get_author()), search_str)
            results = query_google_books(search_str)
            if results:
                logger.info('found isbns from author + title...')
                return results
        return []

    def find_isbn_from_filename(self):
        guessed_meta = self.guess_from_filename()
        if guessed_meta:
            return guessed_meta
        return []

    def fetch_results(self):
        find_functions = [self.find_isbn_in_metadata,
                          self.find_isbn_in_text,
                          self.find_isbn_from_author_title,
                          self.find_isbn_from_filename]
        results = {}
        for f in find_functions:
            for result in f():
                results[result.isbn13] = result
        self.results = list(results.values())
        return self.results


    def _filter_symbols(self, s):
        # type: (str) -> str
        whitelist = string.ascii_letters + string.digits + ' '
        clean_string = ''.join(
            c if c in whitelist else ' ' for c in s)
        return clean_string

    def _clean_filename(self):
        filename, extension = os.path.splitext(self.filename)
        clean_filename = self._filter_symbols(filename)

        return clean_filename

    def guess_from_filename(self):
        clean_filename = self._clean_filename()
        logger.info('looking for %s' % clean_filename)
        results = query_google_books(clean_filename)
        logger.debug('results: %s' % results)
        return results

    def get_isbns_from_text(self):
        pages = self.get_text()
        pages_as_str = '\n'.join(pages)

        isbns = isbnlib.get_isbnlike(pages_as_str, level='normal')

        # print('unprocessed isbns: %s' % isbns)
        canonical_isbns = preprocess_isbns(isbns)

        # print('canonical isbns: %s' % canonical_isbns)
        return canonical_isbns




from sqlalchemy import Column, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship

from .. import db, whooshee

@whooshee.register_model('name')
class Author(db.Model):
    __tablename__ = 'authors'
    __searchable__ = ['name']

    id = Column(Integer, primary_key=True)
    name = Column(String(100))
    books = relationship("Book",
                         secondary="book_author_association",
                         back_populates="authors")

    def __repr__(self):
        return '<{name}>'.format(name=self.name)


import logging

import click
import yaml

from lspace.cli.import_command.add_book_to_db import add_book_to_db
from lspace.cli.import_command.check_if_in_library import check_if_in_library
from lspace.cli.import_command.copy_to_library import _copy_to_library
from lspace.cli.import_command.options import other_choices
from lspace.file_types import get_file_type_object
from lspace.models import Book

logger = logging.getLogger(__name__)

def bold(s):
    return click.style(s, bold=True)

def import_wizard(path, skip_library_check, move):
    click.echo(bold('importing ') + path)

    try:
        f = get_file_type_object(path)
    except Exception as e:
        logger.exception('error reading {path}'.format(path=path), exc_info=True)
        click.secho('error reading {path}'.format(path=path), fg='red')
        return

    if f:
        if not skip_library_check and Book.query.filter_by(md5sum=f.get_md5()).first():
            click.echo(bold('already imported') + ', skipping...')
            return

        isbns_with_metadata = f.fetch_results()

        if len(isbns_with_metadata) == 0:
            click.echo('could not find any isbn or metadata for %s' % f.filename)
            choice = choose_result(f, [])

        else:
            choice = choose_result(f, isbns_with_metadata)

        logger.debug('choice was %s' % choice)

        while choice in list(other_choices.keys()):
            # if choice is one of "other choices", its not one of the results,
            # but one of the strings mapped to functions

            function_that_gets_new_choices = other_choices.get(choice)['function']
            isbns_with_metadata = function_that_gets_new_choices(file_type_object=f,
                                                                 old_choices=isbns_with_metadata,
                                                                 )

            if isbns_with_metadata is not False:
                choice = choose_result(f, isbns_with_metadata)
            else:
                # "skip" is the only function that returns false here
                choice = False

        if choice:
            return _import(f, choice, move)
        else:
            click.echo('skipping %s' % path, color='yellow')

    else:
        # skip, because we have no class to read this file
        click.echo('skipping %s' % path)


def choose_result(file_type_object, isbns_with_metadata):
    if not isbns_with_metadata:
        click.secho('no results found :(', fg='yellow')

    formatted_choices = format_metadata_choices(
        isbns_with_metadata)

    click.echo(''.join(formatted_choices.values()))

    choices = list(formatted_choices.keys())

    ret = click.prompt('choose result for %s' % file_type_object.filename,
                       type=click.Choice(choices))
    if ret in list(other_choices.keys()):
        choice = ret
    else:
        try:
            idx = int(ret) - 1
            choice = isbns_with_metadata[idx]
        except Exception as e:
            logger.exception('cant convert %s to int!' % ret, exc_info=True)
            return False

    return choice


def format_metadata_choices(isbns_with_metadata):
    # type: (Book) -> dict

    formatted_metadata = {
        # 'choice': 'str shown to user'
    }
    for idx, meta in enumerate(isbns_with_metadata):
        logger.info('adding %s' % meta)

        formatted_metadata[str(idx + 1)] = click.style(
            '{index}: {head}\n'.format(index=idx + 1, head=meta.formatted_output_head()),
                                         bold=True)
        formatted_metadata[str(idx + 1)] += meta.formatted_output_details() + '\n'

    for key, val in other_choices.items():
        formatted_metadata[key] = \
            click.style(yaml.dump({
                key: val['explanation']},
                allow_unicode=True), bold=True)

    logger.debug('formatted data is %s' % formatted_metadata)
    return formatted_metadata


def _import(file_type_object, choice, move_file):
    logger.debug('importing %s, %s' % (file_type_object, choice))
    logger.debug(file_type_object)

    similar_books = check_if_in_library(choice)
    if similar_books:
        click.echo(bold('found similar books in library:'))
        for book in similar_books:
            click.echo(bold('{book.authors_names} - {book.title}'.format(book=book)))
            click.echo('isbn: {book.isbn13}'.format(book=book))
            click.echo('{book.path}\n'.format(book=book))
        if not click.confirm('import anyway?'):
            click.echo(bold('skipping %s' % file_type_object.path))
            return

    path_in_library = _copy_to_library(file_type_object, choice, move_file)
    if path_in_library:
        book = add_book_to_db(file_type_object, choice, path_in_library)
        click.secho('imported %s - %s' % (book.authors_names, book.title), fg='green')
        return book
    else:
        click.secho('could not import %s' % file_type_object.path, fg='red')



import click

from lspace.cli.import_command import import_wizard
from lspace.cli import cli
from lspace import db
from lspace.helpers.query import query_db


@cli.command(help='reimport books in library')
@click.argument('query', nargs=-1)
def reimport(query):
    if not query:
        exit()
    results = query_db(query)
    
    for result in results:
        click.echo('\n')
        click.echo('{result.authors_names} - {result.title}'.format(result=result))
        click.echo('{result.full_path}'.format(result=result))
        if click.confirm('reimport this?'):
            # delete without commit to exclude this book from db check
            db.session.delete(result)
            new_result = import_wizard(result.full_path, skip_library_check=True, move=True)
            if new_result:
                db.session.commit()
            else:
                click.secho('no new entry - keeping the old one!')
                db.session.rollback()



from . import cli

from flask_migrate import init, migrate, upgrade


@cli.group()
def db():
    pass


@db.command(name='init')
def _init():
    init()


@db.command(name='migrate')
def _migrate():
    migrate()


@db.command(name='upgrade')
def _upgrade():
    upgrade()



"""empty message

Revision ID: e2bbe7fd9bc7
Revises: b142d6d9e8da
Create Date: 2019-05-20 23:22:16.784309

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e2bbe7fd9bc7'
down_revision = 'b142d6d9e8da'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('meta_cache',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('isbn', sa.String(length=13), nullable=True),
    sa.Column('service', sa.String(length=10), nullable=True),
    sa.Column('date', sa.DateTime(), nullable=True),
    sa.Column('data', sa.Text(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('meta_cache')
    # ### end Alembic commands ###



from flask.cli import AppGroup

cli = AppGroup('cli')

from .import_command import import_command
from .init_command import init
from .list_command import _list
from .tools_command import tools
from .remove_command import remove
from .reimport_command import reimport
from .export_command import export
from .version_command import version

#from .db_command import db



from lspace.cli import remove_command
from tests.cli._cli_test_base import BaseCliTest


class TestListCommand(BaseCliTest):

    def test_remove(self):
        runner = self.app.test_cli_runner()

        result = runner.invoke(remove_command.remove, [self.test_title])

        assert result.exit_code == 0

        assert '{author} - {title}'.format(author=self.test_author_name, title=self.test_title) in result.output



from lspace.cli.init_command import init
from tests.cli._cli_test_base import BaseCliTest


class TestInitCommand(BaseCliTest):

    def test_init_command(self):
        runner = self.app.test_cli_runner()

        # invoke the command directly
        result = runner.invoke(init, [])
        assert result.exit_code == 0
        assert 'written config file to ' in result.output
        assert self.test_dir in result.output



import logging

import click

from lspace import db
from lspace.cli.tools_command import tools
from lspace.models.meta_cache import MetaCache

logger = logging.getLogger(__name__)

@tools.command(help='clear metadata cache')
def _clear_cache():
    num = MetaCache.query.delete()
    db.session.commit()
    click.echo('deleted %s rows' % num)



import os
import shutil
import tempfile
import unittest

from flask_migrate import upgrade

from lspace import create_app
from lspace.models import Book, Author


def get_test_app(test_dir):
    return create_app(app_dir=test_dir)


def get_temp_dir():
    return tempfile.mkdtemp()


class BaseCliTest(unittest.TestCase):

    def setUp(self):

        self.test_dir = get_temp_dir()
        self.app = get_test_app(self.test_dir)
        self.app.config['USER_CONFIG']['library_path'] = os.path.join(self.test_dir, 'library')

        print(self.app.config['USER_CONFIG'])

        self.test_author_name = 'testname'
        self.test_title = 'testtitle'
        self.test_path = '/some/path.ext'

        self.test_author = Author(name=self.test_author_name)
        self.test_book = Book(authors=[self.test_author], title=self.test_title, path=self.test_path)

        with self.app.app_context():
            upgrade()
            from lspace import db
            db.session.add(self.test_author)
            db.session.add(self.test_book)
            db.session.commit()

    def tearDown(self):
        shutil.rmtree(self.test_dir)




import PyPDF2 as pypdf

from ._base import FileTypeBase


class PDF(FileTypeBase):
    extension = '.pdf'

    def __init__(self, path):
        super().__init__(path)

        self.pdf_reader = pypdf.PdfFileReader(self.path)
        self.metadata = self.pdf_reader.getDocumentInfo()
        self.xmp_metadata = self.pdf_reader.getXmpMetadata()

    def get_text(self):
        # type: () -> [str]
        pages = []

        # printing number of pages in pdf file
        for page_idx in range(min(self.pdf_reader.numPages, 30)):
            page = self.pdf_reader.getPage(page_idx)
            try:
                extracted_text = page.extractText()
                pages.append(extracted_text)
            except KeyError:
                # KeyError: '/Contents'
                pass

        # if true, there was at least one page with text
        if True in [bool(page) for page in pages]:
            return pages
        return []

    def get_isbn(self):
        # meta = self.pdf_reader.xmpMetadata

        # never found a pdf with xmp metadata - not sure what the results would look like
        # if meta:
        #    if meta.dc_description:
        #        print(meta.dc_description)
        return None

    def get_author(self):
        if self.metadata:
            return self.metadata.author
        return None

    def get_title(self):
        try:
            if (self.xmp_metadata and
                    self.xmp_metadata.dc_title and
                    self.xmp_metadata.dc_title.get('x-default', False)):
                return self.xmp_metadata.dc_title.get('x-default', False)
        except AttributeError:
            pass

        if self.metadata:
            return self.metadata.title
        return None



import logging

from lspace import whooshee
from lspace.cli.tools_command import tools

logger = logging.getLogger(__name__)


@tools.command(help='rebuild the search index for your library')
def _rebuild_search_index():
    whooshee.reindex()



import logging

import click
import isbnlib
import yaml

from lspace.cli import cli

logger = logging.getLogger(__name__)


@cli.group(help='tools you probably never need... :P')
def tools():
    pass


from lspace.cli.tools_command.clear_cache import _clear_cache
from lspace.cli.tools_command.rebuild_search_index import _rebuild_search_index
from lspace.cli.tools_command.update_paths import _update_paths


@tools.command(help='convert isbn-10 to isbn-13')
@click.argument('dirtyisbn')
def convert_to_isbn13(dirtyisbn):
    click.echo(isbnlib.to_isbn13(dirtyisbn))


@tools.command(help='find metadata for books by words')
@click.argument('words')
def find_meta_by_text(words):
    if isbnlib.is_isbn13:
        click.echo('%s looks like isbn!' % words)
        results = isbnlib.meta(words, service='openl')
    else:
        results = isbnlib.goom(words)
    click.echo(yaml.dump(results))


@tools.command(help='start ipython')
def shell():
    import pdb
    import rlcompleter, readline
    readline.parse_and_bind('tab: complete')

    pdb.Pdb.complete = rlcompleter.Completer(locals()).complete

    breakpoint()



import logging
import os
import shutil
import subprocess

import click
from slugify import slugify

from flask import current_app
from lspace.cli import cli
from lspace.helpers import find_unused_path
from lspace.helpers.query import query_db

logger = logging.getLogger(__name__)


@cli.command(help='export books to a folder - this uses ebook-convert, which is part of calibre')
@click.argument('query', nargs=-1)
@click.argument('export_path')
@click.option('--format')
def export(query, export_path, format):
    if not shutil.which('ebook-convert'):
        click.echo('could not find ebook-convert executable! is calibre installed?')
        return
    export_path = os.path.abspath(os.path.expanduser(export_path))
    click.echo('exporting to %s' % export_path)

    if not os.path.isdir(export_path):
        click.echo('path has to be a folder!')
        return

    if not format.startswith('.'):
        # format and extension should both start with '.'
        format = '.' + format

    results = query_db(query)
    for result in results:
        author_slugs = [slugify(author.name) for author in result.authors]
        if not author_slugs:
            author_slugs = ['UNKNOWN AUTHOR']
        authors_slug = '_'.join(author_slugs)

        authors_str = '_'.join([author.name for author in result.authors])
        title_slug = slugify(result.title)

        if format and format != result.extension:
            target_extension = format
        else:
            target_extension = result.extension

        target_in_export_path = find_unused_path(export_path,
                                                 current_app.config['USER_CONFIG']['file_format'],
                                                 authors_slug,
                                                 title_slug,
                                                 target_extension)
        target_path = os.path.join(export_path, target_in_export_path)
        if not os.path.isdir(os.path.dirname(target_path)):
            os.makedirs(os.path.dirname(target_path))

        if target_extension != result.extension:
            click.echo('converting %s to %s' % (result.extension, format))
            try:
                subprocess.call(
                    ["ebook-convert",
                     result.full_path,
                     target_path
                     ])
            except Exception as e:
                logger.exception('error converting %s' % result.full_path, exc_info=True)
        else:
            click.echo('exporting {authors_str} - {result.title} to {target_path}'.format(authors_str=authors_str,
                                                                                          result=result,
                                                                                          target_path=target_path))
            shutil.copyfile(result.full_path, target_path)
    return



from copy import deepcopy

import click
import yaml

from lspace.file_types import FileTypeBase
from lspace.helpers import preprocess_isbns
from lspace.helpers.query import query_isbn_data, query_google_books
from lspace.models import Book


def peek_function(file_type_object, old_choices, *args, **kwargs):
    click.echo_via_pager(file_type_object.get_text())
    return old_choices


def run_search_function(*args, **kwargs):
    search_string = click.prompt('search string')
    return query_google_books(search_string)


def skip_book_function(*args, **kwargs):
    return False


def lookup_isbn_function(*args, **kwargs):
    isbn_str = click.prompt(
        'specify isbn (without whitespaces, only the number)')
    result = query_isbn_data(isbn_str)
    if result:
        return [result]
    return []


def manual_import(file_type_object, *args, **kwargs):
    # type: (FileTypeBase, [], {}) -> [dict]
    _edit_dict = dict(
        Title='',
        Authors=['', ],
        ISBN='',
        Publisher='',
        Year='',
        Language='',
    )

    edit_dict = deepcopy(_edit_dict)

    while True:
        text = '# import {file_type_object.path}\n'.format(file_type_object=file_type_object)
        text += '# only the title is needed, but you probably want to specify more :)\n\n'
        text += yaml.dump(edit_dict, sort_keys=False)

        result = click.edit(text, require_save=True)
        if not result:
            click.pause('no data! did you save?')
        else:
            result = yaml.load(result, Loader=yaml.FullLoader)
            if not result['Title']:
                choice = ''
                while choice not in ['e', 'r']:
                    choice = click.prompt('title is needed! \n'
                                          'e: edit\n'
                                          'p: ' + other_choices['p']['explanation'] + '\n' +
                                          'r: restart with empty form\n',
                                          type=click.Choice(['e', 'r', 'p']), default='e')
                    if choice == 'e':
                        edit_dict = result

                    elif choice == 'r':
                        edit_dict = deepcopy(_edit_dict)
                    else:
                        other_choices['p']['function'](file_type_object, [])
            else:
                isbn = result.get('ISBN')
                if isbn:
                    isbns = preprocess_isbns([isbn])
                    if len(isbns) == 1:
                        isbn = isbns[0]
                        result['ISBN-13'] = isbn

                return [Book.from_search_result(result, metadata_source='manually added')]


other_choices = {
    'q': dict(
        function=run_search_function,
        explanation='run manual query'),
    'i': dict(
        function=lookup_isbn_function,
        explanation='search by isbn'),
    'p': dict(
        function=peek_function,
        explanation='peek in the text'),
    'm': dict(
        function=manual_import,
        explanation='import manually'),
    's': dict(
        function=skip_book_function,
        explanation='skip this book'),
}



import os
from unittest.mock import MagicMock

from lspace.cli.import_command import copy_to_library
from lspace.file_types import FileTypeBase
from lspace.models import Book
from tests.cli._cli_test_base import BaseCliTest


class TestCopyToLibrary_Move(BaseCliTest):

    def test_copy_to_library(self):
        with self.app.app_context():
            book = Book.query.first()
            old_path = book.path
            new_path = 'new_path'
            absolute_library_path = os.path.join(self.app.config['USER_CONFIG']['library_path'], new_path)

            copy_to_library.move = MagicMock(return_value=None)
            copy_to_library.copyfile = MagicMock(return_value=None)

            copy_to_library.find_unused_path = MagicMock(return_value=new_path)

            file_type_object_mock = FileTypeBase(book.path)
            file_type_object_mock.extension = '.test'

            copy_to_library._copy_to_library(file_type_object_mock, book, move_file=True)
            copy_to_library.move.assert_called_with(old_path, absolute_library_path)
            copy_to_library.copyfile.assert_not_called()

class TestCopyToLibrary_Copy(BaseCliTest):

    def test_copy_to_library_copy(self):
        with self.app.app_context():
            book = Book.query.first()
            old_path = book.path
            new_path = 'new_path'

            absolute_library_path = os.path.join(self.app.config['USER_CONFIG']['library_path'], new_path)

            copy_to_library.move = MagicMock(return_value=None)
            copy_to_library.copyfile = MagicMock(return_value=None)

            copy_to_library.find_unused_path = MagicMock(return_value=new_path)

            file_type_object_mock = FileTypeBase(book.path)
            file_type_object_mock.extension = '.test'

            copy_to_library._copy_to_library(file_type_object_mock, book, move_file=False)
            copy_to_library.move.assert_not_called()

            copy_to_library.copyfile.assert_called_with(old_path, absolute_library_path)



import os

import click

from lspace.cli import cli
from lspace.helpers.query import query_db
from flask import current_app


@cli.command(name='list', help='query your database')
@click.argument('query', nargs=-1)
@click.option('--path', is_flag=True)
@click.option('--details', is_flag=True)
def _list(query, path, details):
    results = query_db(query)

    if path:
        for result in results:
            click.echo(
                result.full_path)
            return

    if details:
        for result in results:
            click.echo('{result.authors_names} - {result.title}'.format(result=result))
            click.echo('{result.full_path}'.format(result=result))
            click.echo('language: {result.language}'.format(result=result))
            click.echo('year: {result.year}'.format(result=result))
            click.echo('isbn: {result.isbn13}'.format(result=result))
            click.echo()
        return

    for result in results:
        click.echo('{result.authors_names} - {result.title}'.format(result=result))




from . import create_app

import os
from flask_migrate import upgrade

def upgrade_db_if_needed(app):
    with app.app_context():
        upgrade()


config_path = os.environ.get('LSPACE_CONFIG', False)

app = create_app(config_path=config_path)


upgrade_db_if_needed(app)

from lspace.cli import cli as cli_group



from setuptools import setup, find_packages

requirements = [
    "alembic==1.0.10",
    "blinker==1.4",
    "Click==7.0",
    "colorama==0.4.1",
    "EbookLib==0.17.1",
    "Flask==1.0.2",
    "flask-marshmallow==0.10.1",
    "Flask-Migrate==2.4.0",
    "Flask-SQLAlchemy==2.4.0",
    "flask-whooshee==0.7.0",
    "html2text==2018.1.9",
    "isbnlib==3.9.8",
    "itsdangerous==1.1.0",
    "Jinja2==2.10.1",
    "lxml==4.3.3",
    "Mako==1.0.10",
    "MarkupSafe==1.1.1",
    "marshmallow==2.19.2",
    "marshmallow-sqlalchemy==0.16.3",
    "PyPDF2==1.26.0",
    "python-dateutil==2.8.0",
    "python-editor==1.0.4",
    "python-slugify==3.0.2",
    "PyYAML==5.1",
    "six==1.12.0",
    "SQLAlchemy==1.3.3",
    "text-unidecode==1.2",
    "Werkzeug==0.15.4",
    "Whoosh==2.7.4",
]

test_requirements = [
    "pytest==4.5.0",
    "pytest-cov==2.7.1",
    "pytest-cover==3.0.0",
    "codecov==2.0.1"
]

dev_requirements = test_requirements + [
    "ipython==7.5.0",
    "doit==0.31.1",
]

setup(
    name='lspace',
    packages=find_packages(),
    include_package_data=True,
    version='0.1.12',
    entry_points='''
        [console_scripts]
        lspace=lspace.app:cli_group
    ''',
    description='a ebook manager built around isbnlib',
    url='https://github.com/puhoy/lspace',
    author='jan',
    author_email='stuff@kwoh.de',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    extras_require={
        'dev': dev_requirements,
        'test': test_requirements
    },
    install_requires=requirements,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        "Intended Audience :: End Users/Desktop",
        "Natural Language :: English",
        "Topic :: Utilities",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3",
    ],
)




import logging

from lspace.file_types import FileTypeBase
from lspace.models import Book

logger = logging.getLogger(__name__)

def add_book_to_db(file_type_object, result, path_in_library):
    # type: (FileTypeBase, Book, str) -> Book
    """

    :param file_type_object: wrapper for the file we want to import
    :param result: chosen metadata result
    :param path_in_library: the path after import
    :return:
    """

    result.md5sum=file_type_object.get_md5()
    result.path=path_in_library

    logger.info('adding book %s' % result.to_dict())

    result.save()

    return result



import logging

import click

from lspace.cli.import_command import _copy_to_library
from lspace.cli.tools_command import tools
from lspace.file_types import get_file_type_object
from lspace.models import Book

logger = logging.getLogger(__name__)


def update_path(book):
    source_path = book.full_path

    file_type_object = get_file_type_object(source_path)

    new_path = _copy_to_library(file_type_object, book, True)

    book.path = new_path
    book.save()

    return new_path


@tools.command(help='update paths after changing file_format setting in config')
def _update_paths():
    books = Book.query.all()
    for book in books:

        source_in_library = book.path
        new_path = update_path(book)

        try:
            click.echo('moved {source} to {new_path}'.format(source=source_in_library, new_path=new_path))
        except Exception as e:
            logger.exception('error moving file', exc_info=True)



import logging

import click

from lspace.cli.import_command._import import import_wizard
from lspace.cli.import_command.add_book_to_db import add_book_to_db
from lspace.cli.import_command.copy_to_library import _copy_to_library

from .. import cli

logger = logging.getLogger(__name__)


@cli.command(name='import', help='import ebooks into your database')
@click.argument('document_path', type=click.Path(exists=True), nargs=-1)
@click.option('--skip-library-check', help='dont check if this file is in the library already', default=False,
              is_flag=True)
@click.option('--move', help='move imported files instead copying', default=False, is_flag=True)
def import_command(document_path, skip_library_check, move):
    for path in document_path:
        import_wizard(path, skip_library_check, move)



"""empty message

Revision ID: 91732d5540b9
Revises: e2bbe7fd9bc7
Create Date: 2019-06-06 22:52:21.057872

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '91732d5540b9'
down_revision = 'e2bbe7fd9bc7'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('books', sa.Column('metadata_source', sa.String(length=20), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('books', 'metadata_source')
    # ### end Alembic commands ###



import logging

from lspace import db
from lspace.models import Book

logger = logging.getLogger(__name__)



def check_if_in_library(result):
    # type: (Book) -> Set[Book]
    """

    :param file_type_object: wrapper for the file we want to import
    :param result: chosen metadata result
    :param path_in_library: the path after import
    :return:
    """

    title = result.title
    isbn13 = result.isbn13
    publisher = result.publisher
    year = result.year
    language = result.language


    books = db.session.query(Book).filter_by(isbn13=isbn13).all()
    if isbn13:
        books += Book.query.filter(Book.title.ilike(title.replace(' ', '%')) ).distinct()
    return set(books)



import os
import click

from lspace.cli import cli
from lspace.helpers.query import query_db
from lspace import db


@cli.command(help='remove books from library')
@click.argument('query', nargs=-1)
def remove(query):
    if not query:
        exit()
    results = query_db(query)
    
    for result in results:        
        click.echo('\n')
        click.echo('{result.authors_names} - {result.title}'.format(result=result))
        click.echo('{result.full_path}'.format(result=result))
        if click.confirm('delete this book from library?'):
            os.unlink(result.full_path)
            db.session.delete(result)
            db.session.commit()



import logging
import os

from shutil import copyfile, move

from lspace.helpers import find_unused_path
from flask import current_app
from lspace.file_types._base import FileTypeBase
from lspace.models import Book

logger = logging.getLogger(__name__)


def _copy_to_library(file_type_object, result, move_file):
    # type: (FileTypeBase, Book, bool) -> str
    """

    :param file_type_object: wrapper for the file we want to import
    :param result: chosen result
    :param move_file: move instead of copy
    :return:
    """
    # prepare the fields for path building

    library_path = current_app.config['USER_CONFIG']['library_path']
    path_in_library = find_unused_path(library_path,
                                       current_app.config['USER_CONFIG']['file_format'],
                                       result.author_names_slug,
                                       result.title_slug,
                                       file_type_object.extension)
    target_path = os.path.join(library_path, path_in_library)

    if not target_path:
        logger.error('could not find a path in the library for %s' %
                     file_type_object.path)
        return False

    if not os.path.isdir(os.path.dirname(target_path)):
        os.makedirs(os.path.dirname(target_path))

    logger.debug('importing to %s' % target_path)
    if not move_file:
        copyfile(file_type_object.path, target_path)
    else:
        if file_type_object.path != target_path:
            move(file_type_object.path, target_path)
        else:
            logger.info('source and target path are the same - skip moving the file')
    file_type_object.path = target_path
    return path_in_library
